""" Contains API Logic"""
from __future__ import annotations

from typing import Any, Dict, List, Optional

from fastapi import BackgroundTasks, FastAPI, Query

from stock_computing_service.database import query_db
from stock_computing_service.helpers import get_curr_time
from stock_computing_service.statistics.model import calculate_statistics
from stock_computing_service.sync import sync

app = FastAPI()


@app.get("/")
async def financial():
    return {"data": "Welcome!"}


# TODO:
# 1. Move below to route module
# 2. Write a Model for below API Arguments
@app.get("/api/financial_data")
async def get_financial_data(
    start_date: Optional[str] = Query(None, description="start date (YYYY-MM-DD)"),
    end_date: Optional[str] = Query(None, description="end date (YYYY-MM-DD)"),
    symbol: Optional[str] = Query("IBM", description="symbol"),
    limit: Optional[int] = Query(5, description="limit of records per page"),
    page: Optional[int] = Query(1, description="page index"),
) -> Dict[str, Any]:
    """
    Endpoint to fetch financial data for symbol based on below parameter
    Args:
        start_date (Optional[str], optional): Start time. Defaults to Query(None, description="start date (YYYY-MM-DD)").
        end_date (Optional[str], optional): End Time. Defaults to Query(None, description="end date (YYYY-MM-DD)").
        symbol (Optional[str], optional): Symbol. Defaults to Query("IBM", description="symbol").
        limit (Optional[int], optional): Limit. Defaults to Query(5, description="limit of records per page").
        page (Optional[int], optional): Page. Defaults to Query(1, description="page index").

    Returns:
        Dict[str, Any]: Response containing financial data
    """
    error: str = ""
    data = None
    try:
        data = await query_db(
            start_date=start_date,
            end_date=end_date,
            symbol=symbol,
            page=page,
            limit=limit,
        )
    except Exception as err:
        error = str(err)
    total_count: int = len(data)
    num_pages: int = (total_count + limit - 1) // limit
    pagination: Dict[str, Any] = {
        "count": total_count,
        "page": page,
        "limit": limit,
        "pages": num_pages,
    }
    result = {"data": data, "pagination": pagination, "info": {"error": error}}
    return result


@app.get("/api/sync")
async def sync_finacial_data(background_tasks: BackgroundTasks) -> Dict[str, str]:
    """
    Triggers sync servoic to fetch financial data and load into database
    Returns:
        Dict[str, str]: Sync Trigger logs
    """
    background_tasks.add_task(sync)
    return {"data": f"Sync Triggered at {get_curr_time()}"}


@app.get("/api/statistics")
async def get_statistics(
    start_date: Optional[str] = Query(None, description="start date (YYYY-MM-DD)"),
    end_date: Optional[str] = Query(None, description="end date (YYYY-MM-DD)"),
    symbol: Optional[str] = Query("IBM", description="symbol"),
) -> Dict[str, Any]:
    """
    API to get statistics of financial data
    Args:
        start_date (str, optional): Start Date. Defaults to Query(None, description="start date (YYYY-MM-DD)").
        end_date (str, optional): End Date. Defaults to Query(None, description="end date (YYYY-MM-DD)").
        symbol (str, optional): Symbol. Defaults to Query(None, description="symbol").

    Returns:
        Dict[str, Any]: Response containing financial data
    """
    error = ""
    output = {}
    try:
        request_data = dict(start_date=start_date, end_date=end_date, symbol=symbol)
        query_res = await query_db(**request_data)
        request_data["results"] = query_res
        output = await calculate_statistics(**request_data)
    except Exception as err:
        error = str(err)
    return {
        "data": output.dict() if type(output) != dict else output,
        "info": {"error": error},
    }



"""
Useful Miscellaneous Utilities which can be reused at multiple places in a Python Code
"""

from __future__ import annotations

import datetime
from typing import Any, Callable, Dict, List

from dateutil import parser


def _map(func: Callable, data: List[Any]) -> List[Any]:
    """Wrapper over python's map fucntion"""
    return [*map(func, data)]


def is_valid_datetime(string):
    try:
        if parser.parse(string):
            return True
    except:
        return False


def sanetize_dict(data):
    symbols = [".", " "]
    output = {}
    for key, val in data.items():
        if isinstance(val, dict):
            val = sanetize_dict(val)
        if is_valid_datetime(key):
            output[key] = val
            continue
        for char in [*range(0, 9)] + symbols:
            key = key.replace(str(char), "")
        output[key] = val
    return output


def get_curr_time():
    now = datetime.datetime.now()
    current_date_time = now.strftime("%Y-%m-%d %H:%M:%S")
    return current_date_time


def db_row_to_dict(row):
    d = {}
    for column in row.__table__.columns:
        d[column.name] = str(getattr(row, column.name))
    return d



from __future__ import annotations

import asyncio
from typing import Any, Dict, List, Optional

import requests

from stock_computing_service.constants import API_KEY, TimeFunctions

scheme: str = "https://"
domain: str = "www.alphavantage.co"


async def get_core_stocks(
    symbols: List[str],
    time_function: Optional[TimeFunctions] = TimeFunctions.TIME_SERIES_WEEKLY,
) -> Dict[str, Any]:
    """
    Fetch Cores Stocks data
    Args:
        symbol (str): Symbol
        time_function (Optional[TimeFunctions], optional): Defaults to TimeFunctions.TIME_SERIES_WEEKLY.
    Returns:
        dict: Finanical Data
    """
    sym_url_map: List[str] = {
        symbol: f"{scheme}{domain}/query?function={time_function}&symbol={symbol}&apikey={API_KEY}"
        for symbol in symbols
    }

    tasks: List[asyncio.Task] = [
        asyncio.ensure_future(call_get_api(url)) for _, url in sym_url_map.items()
    ]
    results = await asyncio.gather(*tasks)
    return {symbol: data for symbol, data in zip(symbols, results)}


async def call_get_api(url: str) -> Dict[str, Any]:
    """Calls Get API"""
    response: requests.Response = requests.get(url, verify=False)
    response.raise_for_status()
    return response.json()



""" Projects constants """
from __future__ import annotations

import os
from enum import Enum

# TODO: Move symbols to DATABASE
SYMBOLS = ["IBM", "AAPL"]
API_KEY = os.environ.get("FINANCIAL_KEY", "demo")

DB_HOST = os.environ.get("DATABASE_HOST", "localhost")


class DatabaseEnum(str, Enum):
    NAME = "stock_computing"
    SERVER = f"postgresql://postgres:postgres@{DB_HOST}:5432/stock_computing"
    HOST = DB_HOST
    USER = "postgres"
    PASSWORD = "postgres"
    DEFAULT = "postgres"
    DRIVER = "postgresql"
    PORT = 5432
    ISOLATION_LEVEL = "AUTOCOMMIT"


# DB = "sqlite:///financial_data.db"


class TimeFunctions(str, Enum):
    TIME_SERIES_WEEKLY = "TIME_SERIES_WEEKLY"


TimeFunctionsMap = {TimeFunctions.TIME_SERIES_WEEKLY: "WeeklyTimeSeries"}



""" Contains Database Logic"""
from __future__ import annotations

import asyncio
from typing import Any, Dict, List, Optional

from sqlalchemy import (
    Column,
    DateTime,
    Float,
    Integer,
    String,
    UniqueConstraint,
    create_engine,
    text,
)

# Create the database if it does not exist
from sqlalchemy.engine.url import URL
from sqlalchemy.exc import IntegrityError, ProgrammingError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Session, sessionmaker
from sqlalchemy.orm.exc import NoResultFound

from stock_computing_service.constants import DatabaseEnum
from stock_computing_service.helpers import _map, db_row_to_dict


def create_db_if_not_exists():
    # Define the PostgreSQL URL for the new database
    url = URL.create(
        drivername=DatabaseEnum.DRIVER.value,
        username=DatabaseEnum.USER.value,
        password=DatabaseEnum.PASSWORD.value,
        host=DatabaseEnum.HOST.value,
        port=DatabaseEnum.PORT.value,
        database=DatabaseEnum.DEFAULT.value,
    )
    # Create a new engine with the 'postgres' database
    engine = create_engine(url)

    # Create a new database if it doesn't already exist
    try:
        with engine.connect().execution_options(
            isolation_level=DatabaseEnum.ISOLATION_LEVEL.value,
        ) as conn:
            conn.execute(text(f"CREATE DATABASE {DatabaseEnum.NAME.value}"))
            print(f"Database '{DatabaseEnum.NAME.value}' created")
    except ProgrammingError:
        print(f"Database '{DatabaseEnum.NAME.value}' already exists")
    engine.dispose()


create_db_if_not_exists()
Base = declarative_base()
engine = create_engine(DatabaseEnum.SERVER.value)


def init_db():
    """
    Initializes database objects
    """
    Session = sessionmaker(bind=engine)
    session = Session()
    Base.metadata.create_all(bind=engine)
    return session


class FinancialDataTable(Base):
    """Model class for Financial Data"""

    __tablename__ = "financial_data"

    id = Column(Integer, primary_key=True)
    symbol = Column(String(50))
    date = Column(DateTime)
    open_price = Column(Float)
    close_price = Column(Float)
    volume = Column(Integer)
    __table_args__ = (UniqueConstraint("symbol", "date"),)


async def data_present(financial_data: Dict[str, str]):
    """Verify if Data is already is present in DB"""
    try:
        query = session.query(FinancialDataTable).filter_by(
            symbol=financial_data.symbol,
            date=financial_data.date,
        )
        query.one()
        return True
    except NoResultFound:
        return False


async def upsert(financial_data: Dict[str, str]):
    """Insert financial data into database"""
    if not await data_present(financial_data):
        financial_data_table = FinancialDataTable(
            symbol=financial_data.symbol,
            date=financial_data.date,
            open_price=financial_data.open_price,
            close_price=financial_data.close_price,
            volume=financial_data.volume,
        )
        session.add(financial_data_table)
        session.commit()


async def upsert_all(
    financial_data_list: List[Dict[str, Any]],
    all_at_once: Optional[bool] = True,
):
    """
    This will add all the instances in the financial_data_table_list to the session
    and commit the changes to the database at once. If the table doesn't exist,
    it will be created automatically based on the
    definition of the FinancialDataTable model.
    Args:
        financial_data (_type_): _description_
    """
    if all_at_once:
        try:
            session.add_all(
                [FinancialDataTable(**data.dict()) for data in financial_data_list],
            )
            session.commit()
        except IntegrityError as err:
            print(f"Failed to insert data: {err}")
        return

    tasks = [asyncio.ensure_future(upsert(data)) for data in financial_data_list]
    await asyncio.gather(*tasks)


async def query_db(
    start_date: Optional[str],
    end_date: Optional[str],
    symbol: Optional[str],
    limit: Optional[int] = None,
    page: Optional[int] = None,
) -> List[Dict[str, str]]:
    """
    Query Database based on filter
    Returns:
        List[Dict[str, str]]: Data fetched from Database
    """
    query = session.query(FinancialDataTable)
    if start_date:
        query = query.filter(FinancialDataTable.date >= start_date)
    if end_date:
        query = query.filter(FinancialDataTable.date <= end_date)
    if symbol:
        query = query.filter(FinancialDataTable.symbol == symbol)
    if limit and page:
        offset = (page - 1) * limit
        query = query.limit(limit).offset(offset)
    return [db_row_to_dict(row) for row in query.all()]


session: Session = init_db()



""" Model class for Statistics """
from __future__ import annotations

from typing import Any, Optional

from pydantic import BaseModel


class Statistics(BaseModel):
    """Model class for managing Stock Statistics"""

    start_date: Optional[str] = ""
    end_date: Optional[str] = ""
    symbol: str
    average_daily_open_price: float
    average_daily_close_price: float
    average_daily_volume: int

    def __init__(__pydantic_self__, **data: Any) -> None:
        super().__init__(**data)


async def calculate_statistics(**data) -> Statistics:
    """Calculates statistics for financial data"""
    finanical_data: List[Any] = data.get("results", [])

    open_prices: List[float] = [float(stock["open_price"]) for stock in finanical_data]
    data["average_daily_open_price"] = sum(open_prices) / len(open_prices)

    close_prices: List[float] = [
        float(stock["close_price"]) for stock in finanical_data
    ]
    data["average_daily_close_price"] = sum(close_prices) / len(close_prices)

    volumes: List[int] = [int(stock["volume"]) for stock in finanical_data]
    data["average_daily_volume"] = sum(volumes) / len(volumes)
    return Statistics(**data)



""" Contains logic to Sync Finanical data into Database"""
from __future__ import annotations

from typing import Dict, List

from stock_computing_service.database import upsert_all
from stock_computing_service.extract.extract import RawStock, extract_stock
from stock_computing_service.transform.transform import FinancialData, transform


async def sync() -> List[FinancialData]:
    """Extract/Transforms/Load the finaical data into DB"""
    extracted_data: Dict[str, RawStock] = await extract_stock()
    transformed_data: List[FinancialData] = transform(extracted_data)
    await upsert_all(transformed_data, all_at_once=False)
    return transformed_data



""" Model class for Transforming Finanical Data"""

from __future__ import annotations

from datetime import datetime
from typing import Dict, List

from pydantic import BaseModel, validate_arguments

from stock_computing_service.extract.extract import RawStock


class FinancialData(BaseModel):
    symbol: str
    date: datetime
    open_price: float
    close_price: float
    volume: int


def _transform(symbol, timeseries_data) -> List[FinancialData]:
    """Helper function for transofrm finanical data"""
    financial_data: List[FinancialData] = [
        FinancialData(
            **dict(
                symbol=symbol,
                date=data.timestamp,
                open_price=float(data.open),
                close_price=float(data.close),
                volume=int(data.volume),
            ),
        )
        for data in timeseries_data
    ]
    return financial_data


@validate_arguments
def transform(stock_data: Dict[str, RawStock]) -> List[FinancialData]:
    """
    Transform RawStockbData in Required format
    Args:
        stock_data (Dict[str, RawStock])
    Returns:
        List[FinancialData]: List of FinancialData
    """
    financial_data: List[FinancialData] = []
    for symbol, raw_stock in stock_data.items():
        financial_data.extend(_transform(symbol, raw_stock.TimeSeries))
    return financial_data



""" Model class for Extracted Data"""
from __future__ import annotations

from typing import Any, Dict, List

from dateutil import parser
from pydantic import BaseModel

from stock_computing_service.constants import SYMBOLS, TimeFunctions, TimeFunctionsMap
from stock_computing_service.extract.api_wrapper import get_core_stocks
from stock_computing_service.helpers import sanetize_dict


class MetaData(BaseModel):
    Information: str
    Symbol: str
    LastRefreshed: str
    TimeZone: str


class TimeSeriesData(BaseModel):
    datetime: str
    open: str
    high: str
    low: str
    close: str
    volume: str

    @property
    def timestamp(self):
        return parser.parse(self.datetime)


class RawStock(BaseModel):
    MetaData: MetaData
    TimeSeries: List[TimeSeriesData]

    def __init__(__pydantic_self__, **data: Any) -> None:
        stock_data = RawStock.adjust_time_function_data(sanetize_dict(data))
        super().__init__(**stock_data)

    @staticmethod
    def adjust_time_function_data(stock_data) -> Dict[str, Any]:
        weekly_data: List[Dict[str, str]] = []
        for time, value in stock_data[
            TimeFunctionsMap[TimeFunctions.TIME_SERIES_WEEKLY]
        ].items():
            value["datetime"] = time
            weekly_data.append(value)
        stock_data["TimeSeries"] = weekly_data
        return stock_data


async def extract_stock() -> Dict[str, RawStock]:
    """Extracts Stocks"""
    extracted_data = sanetize_dict(
        await get_core_stocks(
            symbols=SYMBOLS,
            time_function=TimeFunctions.TIME_SERIES_WEEKLY.value,
        ),
    )
    return {
        symbol: RawStock(**stock_data)
        for symbol, stock_data in extracted_data.items()
        if TimeFunctionsMap[TimeFunctions.TIME_SERIES_WEEKLY] in stock_data
    }
